"use strict";
/**
 * Creates indexes based on the given metadata.
 *
 * @internal
 */
class SchemaCreator {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(schemaBuilder, entityMetadatas) {
        this.schemaBuilder = schemaBuilder;
        this.entityMetadatas = entityMetadatas;
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Creates complete schemas for the given entity metadatas.
     */
    create() {
        const metadatas = this.entityMetadatas;
        return Promise.resolve()
            .then(_ => this.dropForeignKeysForAll(metadatas))
            .then(_ => this.createTablesForAll(metadatas))
            .then(_ => this.updateOldColumnsForAll(metadatas))
            .then(_ => this.dropRemovedColumnsForAll(metadatas))
            .then(_ => this.addNewColumnsForAll(metadatas))
            .then(_ => this.updateExistColumnsForAll(metadatas))
            .then(_ => this.createForeignKeysForAll(metadatas))
            .then(_ => this.updateUniqueKeysForAll(metadatas))
            .then(_ => this.createIndicesForAll(metadatas))
            .then(_ => this.removePrimaryKeyForAll(metadatas))
            .then(_ => { });
    }
    // -------------------------------------------------------------------------
    // Private Methods
    // -------------------------------------------------------------------------
    dropForeignKeysForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.dropForeignKeys(metadata.table, metadata.foreignKeys)));
    }
    createTablesForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.createNewTable(metadata.table, metadata.columns)));
    }
    updateOldColumnsForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.updateOldColumns(metadata.table, metadata.columns)));
    }
    dropRemovedColumnsForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.dropRemovedColumns(metadata.table, metadata.columns)));
    }
    addNewColumnsForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.addNewColumns(metadata.table, metadata.columns)));
    }
    updateExistColumnsForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.updateExistColumns(metadata.table, metadata.columns)));
    }
    createForeignKeysForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.createForeignKeys(metadata.table, metadata.foreignKeys)));
    }
    updateUniqueKeysForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.updateUniqueKeys(metadata.table, metadata.columns)));
    }
    createIndicesForAll(metadatas) {
        return Promise.all(metadatas.map(metadata => this.createIndices(metadata.table, metadata.indices)));
    }
    removePrimaryKeyForAll(metadatas) {
        const queries = metadatas
            .filter(metadata => !metadata.hasPrimaryColumn)
            .map(metadata => this.removePrimaryKey(metadata.table));
        return Promise.all(queries);
    }
    /**
     * Drops all (old) foreign keys that exist in the table, but does not exist in the metadata.
     */
    dropForeignKeys(table, foreignKeys) {
        return this.schemaBuilder.getTableForeignQuery(table.name).then(dbKeys => {
            const dropKeysQueries = dbKeys
                .filter(dbKey => !foreignKeys.find(foreignKey => foreignKey.name === dbKey))
                .map(dbKey => this.schemaBuilder.dropForeignKeyQuery(table.name, dbKey));
            return Promise.all(dropKeysQueries);
        });
    }
    /**
     * Creates a new table if it does not exist.
     */
    createNewTable(table, columns) {
        return this.schemaBuilder.checkIfTableExist(table.name).then(exist => {
            if (!exist)
                return this.schemaBuilder.createTableQuery(table, columns);
            return undefined;
        });
    }
    /**
     * Renames (and updates) all columns that has "oldColumnName".
     */
    updateOldColumns(table, columns) {
        return this.schemaBuilder.getTableColumns(table.name).then(dbColumns => {
            const updates = columns
                .filter(column => !!column.oldColumnName && column.name !== column.oldColumnName)
                .filter(column => dbColumns.indexOf(column.oldColumnName) !== -1)
                .map(column => this.schemaBuilder.changeColumnQuery(table.name, column.oldColumnName, column));
            return Promise.all(updates);
        });
    }
    /**
     * Drops all columns exist (left old) in the table, but does not exist in the metadata.
     */
    dropRemovedColumns(table, columns) {
        return this.schemaBuilder.getTableColumns(table.name).then(dbColumns => {
            const dropColumnQueries = dbColumns
                .filter(dbColumn => !columns.find(column => column.name === dbColumn))
                .map(dbColumn => this.schemaBuilder.dropColumnQuery(table.name, dbColumn));
            return Promise.all(dropColumnQueries);
        });
    }
    /**
     * Adds columns from metadata which does not exist in the table.
     */
    addNewColumns(table, columns) {
        return this.schemaBuilder.getTableColumns(table.name).then(dbColumns => {
            const newColumnQueries = columns
                .filter(column => dbColumns.indexOf(column.name) === -1)
                .map(column => this.schemaBuilder.addColumnQuery(table.name, column));
            return Promise.all(newColumnQueries);
        });
    }
    /**
     * Update all exist columns which metadata has changed.
     */
    updateExistColumns(table, columns) {
        return this.schemaBuilder.getChangedColumns(table.name, columns).then(changedColumns => {
            const updateQueries = changedColumns.map(changedColumn => {
                const column = columns.find(column => column.name === changedColumn.columnName);
                if (!column)
                    throw new Error(`Column ${changedColumn.columnName} was not found in the given columns`);
                return this.schemaBuilder.changeColumnQuery(table.name, column.name, column, changedColumn.hasPrimaryKey);
            });
            return Promise.all(updateQueries);
        });
    }
    /**
     * Creates foreign keys which does not exist in the table yet.
     */
    createForeignKeys(table, foreignKeys) {
        return this.schemaBuilder.getTableForeignQuery(table.name).then(dbKeys => {
            const dropKeysQueries = foreignKeys
                .filter(foreignKey => dbKeys.indexOf(foreignKey.name) === -1)
                .map(foreignKey => this.schemaBuilder.addForeignKeyQuery(foreignKey));
            return Promise.all(dropKeysQueries);
        });
    }
    /**
     * Creates unique keys which are missing in db yet, and drops unique keys which exist in the db,
     * but does not exist in the metadata anymore.
     */
    updateUniqueKeys(table, columns) {
        return this.schemaBuilder.getTableUniqueKeysQuery(table.name).then(dbKeys => {
            // first find metadata columns that should be unique and update them if they are not unique in db
            const addQueries = columns
                .filter(column => column.isUnique)
                .filter(column => dbKeys.indexOf("uk_" + column.name) === -1)
                .map(column => this.schemaBuilder.addUniqueKey(table.name, column.name, "uk_" + column.name));
            // second find columns in db that are unique, however in metadata columns they are not unique
            const dropQueries = columns
                .filter(column => !column.isUnique)
                .filter(column => dbKeys.indexOf("uk_" + column.name) !== -1)
                .map(column => this.schemaBuilder.dropIndex(table.name, "uk_" + column.name));
            return Promise.all([addQueries, dropQueries]);
        });
    }
    /**
     * Creates indices which are missing in db yet, and drops indices which exist in the db,
     * but does not exist in the metadata anymore.
     */
    createIndices(table, compositeIndices) {
        return this.schemaBuilder.getTableIndicesQuery(table.name).then(tableIndices => {
            // drop all indices that exist in the table, but does not exist in the given composite indices
            const dropQueries = tableIndices
                .filter(tableIndex => !compositeIndices.find(compositeIndex => compositeIndex.name === tableIndex.key))
                .map(tableIndex => this.schemaBuilder.dropIndex(table.name, tableIndex.key));
            // then create table indices for all composite indices we have
            const addQueries = compositeIndices
                .filter(compositeIndex => !tableIndices.find(i => i.key === compositeIndex.name))
                .map(compositeIndex => this.schemaBuilder.createIndex(table.name, compositeIndex));
            return Promise.all([dropQueries, addQueries]);
        });
    }
    /**
     * Removes primary key from the table (if it was before and does not exist in the metadata anymore).
     */
    removePrimaryKey(table) {
        // find primary key name in db and remove it, because we don't have (anymore) primary key in the metadata
        return this.schemaBuilder.getPrimaryConstraintName(table.name).then(constraintName => {
            if (constraintName)
                return this.schemaBuilder.dropIndex(table.name, constraintName);
            return undefined;
        });
    }
}
exports.SchemaCreator = SchemaCreator;
//# sourceMappingURL=SchemaCreator.js.map